@@ -79,6 +79,7 @@ MYSQL *mysql_default = NULL;  /* default connection */
    ((mysql_default && strstr(mysql_get_server_info(mysql_default), "maxScale")) ||\
     (getenv("srv")!=NULL && (strcmp(getenv("srv"), "maxscale") == 0 ||\
      strcmp(getenv("srv"), "skysql-ha") == 0)))
+
 #define SKIP_MAXSCALE \
 if (IS_MAXSCALE()) \
 { \
@@ -86,6 +87,17 @@ if (IS_MAXSCALE()) \
   return SKIP; \
 }
 
+#define IS_XPAND()\
+   ((mysql_default && strstr(mysql_get_server_info(mysql_default), "Xpand")) ||\
+    (getenv("srv")!=NULL && strcmp(getenv("srv"), "xpand") == 0))
+
+#define SKIP_XPAND \
+if (IS_XPAND()) \
+{ \
+  diag("test disabled with xpand"); \
+  return SKIP; \
+}
+
 #define SKIP_LOAD_INFILE_DISABLE \
 if (!((mysql->server_capabilities & CLIENT_LOCAL_FILES) &&  \
          (mysql->options.client_flag & CLIENT_LOCAL_FILES))) { \
@@ -617,6 +629,11 @@ void get_envvars() {
   }
   if ((envvar= getenv("MYSQL_TEST_PLUGINDIR")))
     plugindir= envvar;
+
+  if (IS_XPAND())
+  {
+
+  }
 }
 
 MYSQL *my_test_connect(MYSQL *mysql,
@@ -655,7 +672,7 @@ void run_tests(struct my_tests_st *test) {
     total++;
   plan(total);
 
-  /* display TLS stats */
+/* display TLS stats */
   mysql= mysql_init(NULL);
   mysql_ssl_set(mysql, NULL, NULL, NULL, NULL, NULL);
 
